/*
 * Filename: comblock_USB.cpp
 * Date: 3/27/11
 * Author: Alain Zarembowitch / MSS
 * Revision: 0.2
 * Development environment: Microsoft Visual C++ 2010 Express, free downloadable software (recommended over Eclipse/CDT/MinGW)
 * Inheritance: dllmain.cpp 1/2/11
 *
 * Description: utilities to interface with the ComBlock WinUSB driver (rev3)
 *
 * WinUSB Reference document: http://msdn.microsoft.com/en-us/library/ff540174%28v=vs.85%29.aspx#filehandle
 * Other helpful links: http://www.microsoft.com/whdc/connect/usb/default.mspx
 *
 */

// dllmain.cpp : Defines the entry point for the DLL application.
#include "main.h"

int nDevices = 0;
HANDLE hDeviceHandle = NULL; 
BOOL bResult = TRUE;
WINUSB_INTERFACE_HANDLE hWinUSBHandle = NULL; 
unsigned long errNo = 0;
USB_INTERFACE_DESCRIPTOR InterfaceDescriptor;
unsigned char bNumEndpoints = 0;	// number of endpoints in the interface descriptor (2 or 4?)
struct PIPE_ID
{
	UCHAR  PipeInId;
	UCHAR  PipeOutId;
};
WINUSB_PIPE_INFORMATION  Pipe[4];	// ComBlocks support at most 4 pipes (2 for monitoring & control, 2 for high-speed data xfers)
unsigned long bytes_sent = 0;
OVERLAPPED Overlapped;	// overlapped structure
HANDLE hEvent;	// event for overlapped operation
unsigned char pipeId;
unsigned long check = 0;
int mc_in_index;
int mc_out_index;
int data_in_index;
int data_out_index;

//--------------------------------------------------------------------------------
// Method: nActiveDevices
// This method detects how many ComBlocks are connected via USB and ready to communicate.
// driver_rev_no can be 2(now obsolete version. Just checking if the user need to install a new driver) 
// or 3 (current version).
//
// Returns the number of ComBlocks directly connected to this PC via USB.
//
// Error codes: 
// -1 invalid handle value
// -2 invalid driver rev number
//--------------------------------------------------------------------------------
int comblock_USB_nActiveDevices(int driver_rev_no){

	if(driver_rev_no == 2){
		// now obsolete version. Just checking if the user need to install a new driver
		nDevices = nActiveDevices(COMBLOCK_USB_DEVICE_INTERFACE2);
	} else if (driver_rev_no == 3) {
		// current driver version
		nDevices = nActiveDevices(COMBLOCK_USB_DEVICE_INTERFACE3);
	} else {
		return -2;
	}
	return nDevices;
}

//--------------------------------------------------------------------------------
// Method: openHandle
// Get a device handle for the device at the specified index
// The handle will be kept for all other methods until the close method is called.
// Uses usb device driver rev3 (GUID) only (ignores previous driver rev)
// The java application can only communicate with one driver at a time.
// Attempt to open multiple handles will result in closing previous ones.
// 
// This method does more than just creating a device handle and a winusb handle, 
// it also scans the endpoint descriptor to find the correct endpoint locations for
// the two monitoring and control pipes.
// 
// Returns: 0 when everything went OK.
//
// Error codes: 
// -1 device index is out of range. Valid range [0, nDevices-1]
// -2 error getting a device handle
// -3 Could not get a WinUSB interface handle for the specified index
// -4 Error getting interface descriptor
// -5 Error getting endpoint descriptor
// -6 could not find the two M&C endpoint addresses in the endpoint descriptors
//--------------------------------------------------------------------------------
int comblock_USB_openHandle(int deviceIndex){

	int nDevices = 0;	// number of devices matching the device interface ID currently accessible
	unsigned long policy_type = 0;
	unsigned long policy_value = 0;

	// how many ComBlocks are currently reacheable via USB?
	// There could be more than one USB connection between ComBlocks and this PC
	// returns a negative number if error
	nDevices = nActiveDevices(COMBLOCK_USB_DEVICE_INTERFACE3);

	if((deviceIndex < 0) || (deviceIndex >= nDevices)){
		// device index is out of range. Valid range [0, nDevices-1]
		return -1;
	}

	// before opening a new handle, make sure the previous one was closed (Java developer may have forgotten to do so)
	// The java application can only communicate with one driver at a time.
	if(hWinUSBHandle != NULL){
		WinUsb_Free(hWinUSBHandle);
		hWinUSBHandle = NULL;
	}
	if(hDeviceHandle != NULL){
	    CloseHandle(hDeviceHandle);
		hDeviceHandle = NULL;
	}

	// Get a device handle for the device at the specified index deviceIndex
	bResult = GetDeviceHandle(COMBLOCK_USB_DEVICE_INTERFACE3, deviceIndex, &hDeviceHandle);
    if (!bResult) 
    {
		// error getting a device handle
		hDeviceHandle = NULL;
		return -2;
    }

	// Get a WinUSB Interface Handle for the Device
    bResult = WinUsb_Initialize(hDeviceHandle, &hWinUSBHandle);
    if (!bResult) 
    {
	    CloseHandle(hDeviceHandle);
		hWinUSBHandle = NULL;
		hDeviceHandle = NULL;
		return -3;
    }

	// Get Interface Descriptors from the device itself
	// One interface
	memset(&InterfaceDescriptor, 0, sizeof(USB_INTERFACE_DESCRIPTOR));	// fill in with zeros
    bResult = WinUsb_QueryInterfaceSettings(hWinUSBHandle, 0, &InterfaceDescriptor);
	if(!bResult)
    {
        errNo = -4;
        goto done;
    }
	bNumEndpoints = InterfaceDescriptor.bNumEndpoints;	// should be 2 (M&C only) or 4 (M&C and Data)

	// Scan the Endpoints Descriptors from the device itself (2 or 4 endpoints)
	// in order to find ComBlock M&C IN and OUT ports 
	// Reason: depending on the SIE implementation (FPGA, Atmel uC, NXP LPC1700 ARM), M&C streams 
	// can be located at different endpoint locations.
	// General rules: first IN port is the one for ComBlock monitoring and control
	// and first OUT port is the one for ComBlock monitoring and control

	mc_in_index = -1;	// yet unknown ComBlock monitoring and control IN port
	mc_out_index = -1;	// yet unknown ComBlock monitoring and control OUT port
	data_in_index = -1;	// yet unknown ComBlock high-speed data IN port
	data_out_index = -1;	// yet unknown ComBlock high-speed data OUT port
    for (int index = 0; index < bNumEndpoints; index++)
    {
		memset(&Pipe[index],0,sizeof(WINUSB_PIPE_INFORMATION)); // Fills a block of memory with zeros.
        bResult = WinUsb_QueryPipe(hWinUSBHandle, 0, index, &Pipe[index]);
		if(!bResult)
		{
			errNo = -5;
	        goto done;
		}
		//printf("    Endpoint %d: ID=%02X Type=%02X\r\n",index+1,Pipe[index].PipeId,Pipe[index].PipeType);
		if((mc_in_index < 0) && (Pipe[index].PipeId & 0x80)){
			// first IN port is the one for ComBlock monitoring and control
			mc_in_index = index;
		} else if((mc_out_index < 0) && !(Pipe[index].PipeId & 0x80)){
			// first OUT port is the one for ComBlock monitoring and control
			mc_out_index = index;
		} else if(!(mc_in_index < 0) && (Pipe[index].PipeId & 0x80)){
			// second IN port is the one for ComBlock high-speed data
			data_in_index = index;
		} else if(!(mc_out_index < 0) && !(Pipe[index].PipeId & 0x80)){
			// second OUT port is the one for ComBlock high-speed data
			data_out_index = index;
		}
	}

	// did we find the M&C endpoint addresses? if not, this is a deal breaker. 
	if((mc_in_index < 0) || (mc_out_index < 0)){
		// could not find the two M&C endpoint addresses in the endpoint descriptors
		errNo = -6;
		goto done;
	}

	// set pipe policy for each IN pipe
	// (a) timeout
	if(mc_in_index >= 0){
		policy_type = PIPE_TRANSFER_TIMEOUT;
		policy_value = 100;	// 100ms timeout
		bResult = WinUsb_SetPipePolicy(hWinUSBHandle, Pipe[mc_in_index].PipeId, policy_type, 
			sizeof(policy_value), &policy_value);
	}
	if(data_in_index >= 0){
		policy_type = PIPE_TRANSFER_TIMEOUT;
		policy_value = 100;	// 100ms timeout
		bResult = WinUsb_SetPipePolicy(hWinUSBHandle, Pipe[data_in_index].PipeId, policy_type, 
			sizeof(policy_value), &policy_value);
	}

	// set pipe policy for each OUT pipe
	if(mc_out_index >= 0){
		policy_type = PIPE_TRANSFER_TIMEOUT;
		policy_value = 100;	// 100ms timeout
		bResult = WinUsb_SetPipePolicy(hWinUSBHandle, Pipe[mc_out_index].PipeId, policy_type, 
			sizeof(policy_value), &policy_value);
	}
	if(data_out_index >= 0){
		policy_type = PIPE_TRANSFER_TIMEOUT;
		policy_value = 100;	// 100ms timeout
		bResult = WinUsb_SetPipePolicy(hWinUSBHandle, Pipe[data_out_index].PipeId, policy_type, 
			sizeof(policy_value), &policy_value);
	}

	return 0;

done:
    CloseHandle(hDeviceHandle);
    WinUsb_Free(hWinUSBHandle);
	hWinUSBHandle = NULL;
	hDeviceHandle = NULL;
	return errNo;
}

//--------------------------------------------------------------------------------
// Method: closeHandle
// Close the existing handles (file + winusb)
// Returns: n/a
//
// Error codes: n/a
//--------------------------------------------------------------------------------
void comblock_USB_closeHandle(void){

	if(hWinUSBHandle != NULL){
		WinUsb_Free(hWinUSBHandle);
		hWinUSBHandle = NULL;
	}
	if(hDeviceHandle != NULL){
	    CloseHandle(hDeviceHandle);
		hDeviceHandle = NULL;
	}
	return;
}

//--------------------------------------------------------------------------------
// Method: Read
// 
// Reads from data stream 1 (M&C) or 2 (high-speed data)
// 
// If everything goes well, the number of bytes read (IN from device) will be returned.
// Otherwise, an error code will be returned. Blocking call, limited by timeout (see pipe policy above)
//
// Returns: number of bytes read (could be less than requested)
//
// Error codes: 
// -10 no handle to the driver is present. User needs to call openHandle() first.
// -11 Missing a pointer to the buffer to hold the read data
// -13 Can't read a negative number of bytes
// -14 Could not detect M&C IN endpoint in descriptors
// -15 Could not detect high-speed data IN endpoint in descriptors
// -16 invalid dataStream input selection
// -18 Some other error reading from USB device
// -19 timeout reading
//--------------------------------------------------------------------------------
int comblock_USB_Read(int dataStream, int noOfBytesToRead, 
	char * bufferWithValuesToRead){

	int retval = 0;
	unsigned long bytes_read;

	// various input checks
	if((hWinUSBHandle == NULL) || (hDeviceHandle == NULL)){
		// no handle to the driver is present. User needs to call openHandle() first.
		return -10;
	}
	if(bufferWithValuesToRead == NULL){
		// Missing a pointer to the buffer to hold the read data
		return -11;
	}
	if(noOfBytesToRead < 0){
		// Can't read a negative number of bytes
		return -13;
	}
	if(dataStream == 1){
		// read from M&C IN pipe
		if(mc_in_index < 0){
			// Could not detect M&C IN endpoint in descriptors
			return  -14;
		} else {
			pipeId = Pipe[mc_in_index].PipeId;
		}
	} else if (dataStream == 2){
		// read from high-speed data IN pipe
		if (data_in_index < 0){
			// Could not detect high-speed data IN endpoint in descriptors
			return  -15;
		} else {
			pipeId = Pipe[data_in_index].PipeId;
		}
	} else {
		// invalid dataStream input selection
		return  -16;
	}

	// request a read from ComBlock. Synchronous (i.e. blocking).  
	bytes_read = 0;
	retval = 0;
	bResult = WinUsb_ReadPipe(hWinUSBHandle, pipeId, (unsigned char *)(bufferWithValuesToRead), noOfBytesToRead, &bytes_read, 0);
    if(!bResult){
		errNo = GetLastError();
		if(errNo == ERROR_SEM_TIMEOUT){
			// The read operation initiated by WinUsb_ReadPipe in the USB stack timed out before the operation could be completed. 
			retval = -19;
		} else {
			// Some other error reading from USB device
			retval = -18;
		}
    } else {
		// Got it!
		retval = bytes_read;
	}

	return retval;
}

//--------------------------------------------------------------------------------
// Method: Write
// 
// Write to data stream 1 (M&C) or 2 (high-speed data)
// 
// Returns: number of bytes sent. Blocking calls, limited by timeout (see pipe policy above)
//
// Error codes: 
// -10 no handle to the driver is present. User needs to call openHandle() first.
// -11 Missing a pointer to the buffer holding the data to be written
// -13 Can't write a negative number of bytes
// -14 Could not detect M&C OUT endpoint in descriptors
// -15 Could not detect high-speed data OUT endpoint in descriptors
// -16 invalid dataStream input selection
// -18 Some other error writing to USB device
// -19 timeout writing
//--------------------------------------------------------------------------------
int comblock_USB_Write(int dataStream, int noOfBytesToWrite, 
	char * bufferWithValuesToWrite){

	int retval = 0;

	// various input checks
	if((hWinUSBHandle == NULL) || (hDeviceHandle == NULL)){
		// no handle to the driver is present. User needs to call openHandle() first.
		return -10;
	}
	if(bufferWithValuesToWrite == NULL){
		// Missing a pointer to the buffer holding the data to be written
		return -11;
	}
	if(noOfBytesToWrite < 0){
		// Can't write a negative number of bytes
		return -13;
	}
	if(dataStream == 1){
		// write to M&C OUT pipe
		if(mc_out_index < 0){
			// Could not detect M&C OUT endpoint in descriptors
			return  -14;
		} else {
			pipeId = Pipe[mc_out_index].PipeId;
		}
	} else if (dataStream == 2){
		// write to high-speed data OUT pipe
		if (data_out_index < 0){
			// Could not detect high-speed data OUT endpoint in descriptors
			return  -15;
		} else {
			pipeId = Pipe[data_out_index].PipeId;
		}
	} else {
		// invalid dataStream input selection
		return  -16;
	}

	// request a write from ComBlock. Synchronous (i.e. blocking).  
	bytes_sent = 0;
	retval = 0;
	bResult = WinUsb_WritePipe(hWinUSBHandle, pipeId, (unsigned char *)(bufferWithValuesToWrite), noOfBytesToWrite, &bytes_sent, 0);
    if(!bResult){
		errNo = GetLastError();
		if(errNo == ERROR_SEM_TIMEOUT){
			// The write operation initiated by WinUsb_WritePipe in the USB stack timed out before the operation could be completed. 
			retval = -19;
		} else {
			// Some other error writing from USB device
			retval = -18;
		}
    } else {
		// Got it!
		retval = bytes_sent;
	}
	return retval;
}
//--------------------------------------------------------------------------------
// checks how many active devices match the device interface ID
int nActiveDevices(GUID guidDeviceInterface){
    HDEVINFO hDeviceInfo;
    SP_DEVINFO_DATA DeviceInfoData;
	int i; 
    BOOL bResult;
	long errNo = 0;

	if (guidDeviceInterface==GUID_NULL)
    {
        return 0;
    }
    // Get information about all the installed devices for the specified
    // device interface class.
    hDeviceInfo = SetupDiGetClassDevs( 
        &guidDeviceInterface,
        NULL, 
        NULL,
        DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

    if (hDeviceInfo == INVALID_HANDLE_VALUE) 
    { 
		errNo = GetLastError();	// Error SetupDiGetClassDevs
		return -1;	
    }

    //Enumerate all the device interfaces in the device information set.
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
	for(i = 0; ; i++){
		bResult = SetupDiEnumDeviceInfo(hDeviceInfo, i, &DeviceInfoData);
		if(!bResult)
			break;
	}
	return i;

}

//--------------------------------------------------------------------------------
// Get a device handle for the device at the specified index
BOOL GetDeviceHandle (GUID guidDeviceInterface, int deviceIndex, PHANDLE hDeviceHandle){
    BOOL bResult = TRUE;
    HDEVINFO hDeviceInfo;
    SP_DEVINFO_DATA DeviceInfoData;
    SP_DEVICE_INTERFACE_DATA deviceInterfaceData;
    PSP_DEVICE_INTERFACE_DETAIL_DATA pInterfaceDetailData = NULL;
    ULONG requiredLength=0;
    LPTSTR lpDevicePath = NULL;

    if (guidDeviceInterface==GUID_NULL)
    {
        return FALSE;
    }

    // Get information about all the installed devices for the specified
    // device interface class.
    hDeviceInfo = SetupDiGetClassDevs( 
        &guidDeviceInterface,
        NULL, 
        NULL,
        DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

    if (hDeviceInfo == INVALID_HANDLE_VALUE) 
    { 
        // ERROR 
        printf("Error SetupDiGetClassDevs: %d.\n", GetLastError());
        goto done;
    }

    //Enumerate all the device interfaces in the device information set.
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);

	SetupDiEnumDeviceInfo(hDeviceInfo, deviceIndex, &DeviceInfoData);
	// The SetupDiEnumDeviceInfo function returns a SP_DEVINFO_DATA structure that specifies 
	// a device information element in a device information set

	deviceInterfaceData.cbSize = sizeof(SP_INTERFACE_DEVICE_DATA);

    //Get information about the device interface.
    bResult = SetupDiEnumDeviceInterfaces( 
        hDeviceInfo,
        &DeviceInfoData,
        &guidDeviceInterface,
        0, 
        &deviceInterfaceData);

    //Check for some other error
    if (!bResult) 
    {
        printf("Error SetupDiEnumDeviceInterfaces: %d.\n", GetLastError());
        goto done;
    }

    //Interface data is returned in SP_DEVICE_INTERFACE_DETAIL_DATA
    //which we need to allocate, so we have to call this function twice.
    //First to get the size so that we know how much to allocate
    //Second, the actual call with the allocated buffer
        
    bResult = SetupDiGetDeviceInterfaceDetail(
        hDeviceInfo,
        &deviceInterfaceData,
        NULL, 0,
        &requiredLength,
        NULL);


    //Check for some other error
    if (!bResult) 
    {
        if ((ERROR_INSUFFICIENT_BUFFER==GetLastError()) && (requiredLength>0))
        {
            //we got the size, allocate buffer
            pInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)LocalAlloc(LPTR, requiredLength);
                
            if (!pInterfaceDetailData) 
            { 
                // ERROR 
                printf("Error allocating memory for the device detail buffer.\n");
                goto done;
            }
        }
        else
        {
            printf("Error SetupDiEnumDeviceInterfaces: %d.\n", GetLastError());
            goto done;
        }
    }

    //get the interface detailed data
    pInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    //Now call it with the correct size and allocated buffer
    bResult = SetupDiGetDeviceInterfaceDetail(
            hDeviceInfo,
            &deviceInterfaceData,
            pInterfaceDetailData,
            requiredLength,
            NULL,
            &DeviceInfoData);
        
    //Check for some other error
    if (!bResult) 
    {
        printf("Error SetupDiGetDeviceInterfaceDetail: %d.\n", GetLastError());
        goto done;
    }

    //copy device path
                
    size_t nLength = wcslen (pInterfaceDetailData->DevicePath) + 1;  
    lpDevicePath = (TCHAR *) LocalAlloc (LPTR, nLength * sizeof(TCHAR));
    StringCchCopy(lpDevicePath, nLength, pInterfaceDetailData->DevicePath);
    lpDevicePath[nLength-1] = 0;	// string termination
                        
    printf("Device path:  %s\n", lpDevicePath);


    if (!lpDevicePath)
    {
        //Error.
        printf("Error %d.", GetLastError());
        goto done;
    }

    //Open the device
    *hDeviceHandle = CreateFile (
        lpDevicePath,
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_OVERLAPPED,
        NULL);

    if (*hDeviceHandle == INVALID_HANDLE_VALUE)
    {
        //Error.
        printf("Error %d.", GetLastError());
        goto done;
    }



done:
    LocalFree(lpDevicePath);
    LocalFree(pInterfaceDetailData);    
    bResult = SetupDiDestroyDeviceInfoList(hDeviceInfo);
    
    return bResult;
}

//--------------------------------------------------------------------------------
