% MATLAB program to sort data captured using 
% the Com-8002 and the Com-3002 modules into 
% I and Q samples. 20-bit input format is
% assumed. The file must have been saved 
% in binary format (decimal files are
% NOT supported in this code).
% When the code has finished executing, the 
% I-samples will be stored in the variable 
% iSamples and the Q-samples will be stored 
% in the variable qSamples.

% Note:
% The purpose of this code is to demonstrate
% in an easy manner how the desired data
% format is obtained from the captured data
% format. The code can easily be changed 
% to work with other ComBlock modules and 
% other input width options. 


% Replace 'file_name_including_full_path'
% with the full path and file name for the 
% file where the binary data is stored. 
% For example:
% fid = fopen('c:\My Documents\Captured_Data.txt','r');
fid = fopen('file_name_including_full_path','r'); 


% Read the content of the file into an array.
data = fread(fid,inf,'uchar'); 


% Find out how many samples are in the array.
[dataHeight garbage] = size(data); 


% Find out how big the matrix holding the
% 20-bit wide data needs to be. 
binArrayHeight = ceil(dataHeight/2.5); 


% Pre-initialize the matrix holding 20-bit data 
% for faster program execution. 
binArray = zeros(binArrayHeight,20); 


% Initialize the index counter.
rowCounter = 1; 


% Start the conversion from 8 bits to 20 bits.
% Note: the de2bi function places the MSb 
% on the rightmost position. Therefore, 
% use it with fliplr to get the MSb in the
% leftmost position.
for i = 1:2:binArrayHeight 
% Store I sample 
binArray(i,1:8) = fliplr(de2bi(data(rowCounter),8)); % Store bits 19-12
rowCounter = rowCounter + 1;

binArray(i,9:16) = fliplr(de2bi(data(rowCounter),8)); % Store bits 11-4
rowCounter = rowCounter + 1;

binaryWord = fliplr(de2bi(data(rowCounter),8)); % A temporary array
binArray(i,17:20) = binaryWord(1:4); % Store bits 3-0


% Store Q sample
binArray(i+1,1:4) = binaryWord(5:8); % Store bits 19-16
rowCounter = rowCounter + 1;

binArray(i+1,5:12) = fliplr(de2bi(data(rowCounter),8)); % Store bits 15-8
rowCounter = rowCounter + 1;

binArray(i+1,13:20) = fliplr(de2bi(data(rowCounter),8)); % Store bits 7-0
rowCounter = rowCounter + 1;
end 


% Compensate for the fact that the pinouts are
% reversed (MSb on Com 8002 input connector 
% corresponds to LSb on Com 3002 output connector). 
binArray(:, 1:10) = binArray(:, 10:-1:1); % Reverse I
binArray(:, 11:20) = binArray(:, 20:-1:11); % Reverse Q


% Pre-initialize the arrays for faster program execution.
iSamples = zeros(binArrayHeight, 1); 
qSamples = zeros(binArrayHeight, 1); 


% Convert the 10-bit binary numbers into decimal numbers
% in the range 0-1023. 
for i = 1:binArrayHeight 
iSamples(i,1) = bi2de(fliplr(binArray(i, 11:20)));
qSamples(i,1) = bi2de(fliplr(binArray(i, 1:10)));
end 


% Plot the I and the Q samples.
figure(1) 
plot(iSamples); 
title('I Samples') 
figure(2) 
plot(qSamples); 
title('Q Samples')